/** fMSX: portable MSX emulator ******************************/
/**                                                         **/
/**                        Colors.c                         **/
/**                                                         **/
/** This file contains functions to handle allocation       **/
/** and freeing colours on any X-terminal.                  **/
/** Used by my Improved X-drivers (Unix.c and Refresh.h).   **/
/**                                                         **/
/** Copyright (C) Arnold Metselaar 1999-2001                **/
/**     You are not allowed to distribute this software     **/
/**     commercially. Please, notify me, if you make any    **/
/**     changes to this file.                               **/
/*************************************************************/


/** Standard Unix/X #includes ********************************/
#include <stdio.h>
#include <stdlib.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xatom.h>
#include <X11/keysym.h>

#include "MSX.h"
#include "AnyX.h"
/* not in AnyX.h because it needs X11-includes, that are not in Refresh.c */
extern XVisualInfo VisInfo;

typedef struct {
    int refcnt, cntdwn;
    Status alocd;
    byte ci_pal[3];
    XID colno;
} Colorinfo;

/* global linkage: */
int InitColor(int, int, Colormap*);
void TrashColor(void);
ColTyp Swap(XID X);
XID UnSwap(ColTyp X);

Colorinterface Colors;
ColTyp XPal[16],XPal8[256];
XID white,black;
Bool lsbfirst;
int PalUsed, PalOut, Used8 , IsPal8; 

/* file linkage */

/* prototypes */
static ColTyp GetBorW(byte *, int *, byte *);
static ColTyp GetColStGr(byte *, int *, byte *);
static ColTyp GetColStCo(byte *, int *, byte *);
static ColTyp GetColMut(byte *, int *, byte *);
static ColTyp GetColTrCo(byte *, int *, byte *);
static ColTyp (*pGetColor)(byte *, int *, byte *);
static ColTyp UseXPal8 (byte *pal2p);
static ColTyp GetTrCol2p (byte *pal2p);

static int ColCmp(byte *palx, byte *paly);
static int CompareY(byte * palx, byte * paly);
static int (*pPalCmp)(byte *, byte *);
static int SearchColor(byte * PalReq);
static int ColDist(byte * pal1, byte *pal2);
static int GrayDist(byte * pal1, byte *pal2);
static int (*pPalDist)(byte * pal1, byte *pal2);
static void SetColorStat(byte N,byte R,byte G,byte B);
static void SetColorMut(byte N,byte R,byte G,byte B);
static void (*pSetColor)(byte N,byte R,byte G,byte B);
void SetColor(byte N,byte R,byte G,byte B) { pSetColor(N, R, G, B);}

static void TakeColors(int ColSet);
static void LeaveColors(int ColSet);
static void TakeColorsScr8(void);
static void LeaveColorsScr8(void);
static void RefreshColors(void);
static void UnGetColor(int *, byte *);

static int Y(byte* pal);
static int InitTable(int PrivCMap, int depth); /* for static/direct color */
static int StCoTable(int depth);
static int ShiftFor(long msk);

static void nop(void) {}
static void nop_int(int dummy) {}

static byte Xtab[4];
static byte Pal[16][3]; /* wat we hebben */
static byte Palette[16][3]; /* wat we zouden willen */
static byte nulpal[3]={0,0,0};
static byte  nopal[3]={'i','n','v'};
static int ncol, nprealloc;
static Colormap CMap;
static int CMapIsMine;
static int MutCMap; /* flag that indicates whether the colormap is mutable */


static ColTyp *GrayRamp; /* for use with static gray */
static ColTyp *StatCols; /* for use with static color */
static long mask[3]; /* only used for true color visuals */
static int shift[3]; /* shifts needed to align color values (32 bit)
			with their masks */

/* for use with writable visuals */
static int Ncoltable;
static Colorinfo *coltable;
int alind[16],alind8[256]; /* indices to coltable */

/* set up tables and stuff */
int
InitColor(int do_install, int do_mono, Colormap * p)
{
  int screen_num,i;
  char *order;
  ColTyp T;

  screen_num=VisInfo.screen;
  if (do_install) {
      CMap=XCreateColormap(Dsp, RootWindow(Dsp,screen_num), 
			   VisInfo.visual, AllocNone);
      CMapIsMine=1;
  } 
  else 
    CMap=DefaultColormap(Dsp,screen_num);
  if (Verbose & 1) 
      printf("using %s colormap (0x%lx)\n", 
	     CMapIsMine?"private":"default", CMap);
  *p=CMap;

  black=XBlackPixel(Dsp,screen_num);
  white=XWhitePixel(Dsp,screen_num);

  order="";
  if (do_mono)
    {    
      order=((lsbfirst=(BitmapBitOrder(Dsp)==LSBFirst)) 
	     ? "(lsb first)" : "(msb first)"); 
      printf("    depth=%d, black and white mode %s\n", VisInfo.depth,order);
      bitpix=1;
      pGetColor=GetBorW;
      pPalCmp=CompareY;
      pPalDist=GrayDist;
      MutCMap=0;
    }
  else {
    if (!ChooseFormat())  return 0;
    if (bitpix>8)
      {
	order=(lsbfirst=(ImageByteOrder(Dsp)==LSBFirst)) 
	    ? "(LSB first)" 
	    : "(MSB first)";
	/* how to swap the bytes ? */
	T.ct_xid = lsbfirst ? 0x03020100L : 0x00010203L ;
	if (ImageByteOrder(Dsp)==MSBFirst)
	  for (i=bitpix ; i<32 ; i+=8)
	    T.ct_xid=(T.ct_xid>>8)+((T.ct_xid&0xFF)<<24);
	/* if T were represented by the bytes 0 1 2 3 
	   then no swapping would be needed.. */
	for (i=0;i<4;i++) 
	  Xtab[T.ct_Byte[i]]=i;
	/*  Xtab[j]==i means that byte i in the representation of ColTyp used
	    by the machine on which fmsx is running, corresponds to byte j in
	    the representation of ColTyp in the XImage 
	*/
      } 
    if (bitpix<8)
      order=(lsbfirst=(BitmapBitOrder(Dsp)==LSBFirst)) 
	? "(lsb first)" : "(msb first)"; 
    if (Verbose & 1) 
      printf("    depth=%d, %d bit%s per pixel %s\n", VisInfo.depth,bitpix,
	     (bitpix==1)?"":"s",order);
    if ((VisInfo.depth<=0)||(bitpix<VisInfo.depth))
      {
	if (Verbose & 1)
	  printf("    This is illegal, exiting\n");
	return 0;
      }
    if ((VisInfo.class==StaticGray)||(VisInfo.class==GrayScale))
      {
	pPalCmp=CompareY;
	pPalDist=GrayDist;
      }
    else
      {
	pPalDist=ColDist;
	pPalCmp=ColCmp;
      }
    switch (VisInfo.class) 
      {
      case StaticGray: 
	{
	  XColor gray;
	  int j;
	  GrayRamp=malloc(sizeof(ColTyp)*99);
	  if (!GrayRamp)
	    return 0;
	  
	  gray.flags=DoRed|DoGreen|DoBlue;
	  for( j=0 ; j<99 ; j++) {
	    gray.red=gray.green=gray.blue=(long)j*65535/98; 
	    if (XAllocColor(Dsp,CMap,&gray))
	      GrayRamp[j]=Swap(gray.pixel);
	    else
	      GrayRamp[j]=(j==0)? Swap(black): GrayRamp[j-1];
	  }
	  pGetColor=GetColStGr;
	}
	MutCMap=0;
	break;
	  
      case GrayScale:
	InitTable(CMapIsMine,VisInfo.depth);
	pGetColor=GetColMut;
	MutCMap=1;
	break;

      case StaticColor:
	if (! StCoTable(VisInfo.depth)) return 0;
	pGetColor=GetColStCo;
	MutCMap=0;
	break;

      case PseudoColor:
      case DirectColor: /* don't use the special structure (for now) */
	InitTable(CMapIsMine,VisInfo.depth);
	pGetColor=GetColMut;
	MutCMap=1;
	break;

      case TrueColor: 
	{
	  int j;
	  mask[0]=VisInfo.red_mask;
	  mask[1]=VisInfo.green_mask;
	  mask[2]=VisInfo.blue_mask;
	  for( j=0 ; j<3 ; j++)
	    shift[j]=ShiftFor(mask[j]);
	  pGetColor=GetColTrCo;
	}
	MutCMap=0;
      break;

      default:
	printf("Unknown visual class...exiting\n");
	return 0;
      }
  }
  if (MutCMap) { /* keep track of allocations etc */
    int j;

    for (j=0 ; j<16 ; j++) {
      alind[j]=-1;
      memcpy(Palette[j], nulpal, 3);
      memcpy(Pal[j], nopal, 3);
    }
    for (j=0 ; j<256 ; ) 
      alind8[j++]=-1; /* nothing allocated */  
    PalOut=0xffff;
    PalUsed=Used8=IsPal8=0;

    Colors.Take=TakeColors;
    Colors.Leave=LeaveColors;
    Colors.TakeScr8=TakeColorsScr8;
    Colors.LeaveScr8=LeaveColorsScr8;
    Colors.Refresh=RefreshColors;
    pSetColor=SetColorMut;
  } else {
    int r,g,b, index;
    byte palet[3],dummy[3];

    /* fill in XPal8 once */
    for (r=0 ; r<8 ; r++) {
      palet[0]=r<<5;
      for (g=0 ; g<8 ; g++) {
	palet[1]=g<<5;
	for (b=0 ; b<4 ; b++) {
	  palet[2]= (b==3) ? 7<<5 : b<<6;
	  index=(r<<2)|(g<<5)|b;
	  XPal8[index]=pGetColor(palet,&alind8[index],dummy);
	}
      }
    }
    IsPal8=1;
    /* little to do while running: */
    Colors.Take=nop_int;
    Colors.Leave=nop_int;
    Colors.TakeScr8=nop;
    Colors.LeaveScr8=nop;
    Colors.Refresh=nop;
    pSetColor=SetColorStat;
  }
  Colors.GetCol2p=(VisInfo.class==TrueColor) ? GetTrCol2p: UseXPal8;
  choosefuncs(lsbfirst,bitpix);
  setwidth(wide);

  return 1;
}

static int 
InitTable(int PrivCMap, int depth) /* for pseudo/direct color */
{
  int j;
  Ncoltable=(depth <= 9)? (1<<depth) : 512;
  coltable=malloc(sizeof(Colorinfo)*Ncoltable);
  if (!coltable) return 0;
  
  if(! PrivCMap) {
    black=XBlackPixel(Dsp,VisInfo.screen);
    coltable[0].refcnt=1;
    coltable[0].alocd=1;
    memcpy(coltable[0].ci_pal, nulpal, 3);
    coltable[0].colno=black;
    
    white=XWhitePixel(Dsp,VisInfo.screen);
    coltable[1].refcnt=1;
    coltable[1].alocd=1;
    memcpy(coltable[1].ci_pal,"\340\340\340",3);
    coltable[1].colno=white;
    
    nprealloc=2;
  } else {
    nprealloc=0;
  }
  for (j=nprealloc ; j < Ncoltable ; j++) {
    coltable[j].refcnt=0;
    coltable[j].alocd=0;
  }

  if (ncol==0) { 
    if (PrivCMap) 
      ncol=Ncoltable;
    else
      ncol=((depth>8)||(depth<6))? Ncoltable:64;
  }
  return 1;
}

static int  
StCoTable(int depth) /* StaticColor: make a look-up table once */ 
{
  int r,g,b,res;

  StatCols=malloc(512*sizeof(ColTyp));
  if (!StatCols) return 0;

  for(r=0 ; r<8 ; r++)
    for(g=0 ; g<8 ; g++)
      for(b=0 ; b<8 ; b++) {
	XColor Color;
	int i;

	i= (r<<6) | (g<<3) | b;
	Color.red  =r*65535/7;
	Color.green=g*65535/7;
	Color.blue =b*65535/7;

	res=XAllocColor(Dsp,CMap,&Color);
	if (res)
	  StatCols[i]=Swap(Color.pixel);
	else
	  {
	    printf("shouldn't happen: Color Faillure: R:%d G:%d B:%d \n",r,g,b);
	    StatCols[i]=Swap(black);
	  }
      }
  return 1;
}


static int 
Y(byte * pal)
{ 
  int r,g,b;

  r=pal[0]>>5;
  g=pal[1]>>5;
  b=pal[2]>>5;
/* This is an approximation to the following formula:
 *  Y = 0.212671 * R + 0.715160 * G + 0.072169 * B;  */
  return (3*r + 10*g + 1*b);
}

/* preswap color no's for faster generation of images */
ColTyp 
Swap(XID Color)
{
  int i; 
  ColTyp retval;

  if (bitpix <= 8)
    retval.ct_Byte[0]=Color;
  else 
    for(i=0; i<4; i++) 
      retval.ct_Byte[i]=((ColTyp)Color).ct_Byte[Xtab[i]];
  return retval;
}

/* Undoes byte-swapping done by Swap() */
XID 
UnSwap(ColTyp X)
{
  ColTyp R;
  register byte i;
  
  if (bitpix <= 8) 
    R.ct_xid=X.ct_Byte[0];
  else 
    for(i=0; i<4; i++) 
      R.ct_Byte[Xtab[i]]=X.ct_Byte[i];
  return R.ct_xid;
}

static int 
ShiftFor(long mask)
{
  int i;
  long m;

  for ( i=0,m=0x80000000L ; ! (m&mask) ; i++)
    m>>=1;
  return i;
}

/* The GetCol*() return the closest match for the color described in
   Palet. After execution *alind may be set to the index in some table and
   the palet from the call to GetCol*() that led to the original
   allocation of the color is copied to pal2 The output is
   byte-swapped for faster building of pixmaps. 
*/

static ColTyp 
GetColStGr(byte *Palet, int *allind, byte *pal2)
{
  memcpy(pal2,Palet,3);
  return GrayRamp[Y(Palet)];
}

static ColTyp
GetBorW (byte *Palet, int *allind, byte *pal2)
{
  ColTyp R;
  R.ct_Byte[0]=(Y(Palet)>48) ? 1 : 0;
  return R;
}

static ColTyp 
GetColStCo(byte *Palet, int *allind, byte *pal2)
{
  int r,g,b;

  r=(int)(Palet[0])&0xe0;
  g=(int)(Palet[1])&0xe0;
  b=(int)(Palet[2])&0xe0;
  memcpy(pal2,Palet,3);

  return StatCols[(r<<1)|(g>>2)|(b>>5)];
}


static ColTyp 
GetColMut(byte *Palet, int *allind, byte *pal2)
{
  int i,j, index; Status res;
  int mindist, imin;

  memcpy(pal2,Palet,3);

  /* check if this color was previously allocated */
  index = SearchColor(Palet);
  if (index >=0) goto found;

  /* search for free entry in coltable */
  for (i=nprealloc ; (coltable[i].alocd) && (i<ncol) ; i++)
    ;
  if (i<ncol) {
    XColor Color;
    
    Color.flags=DoRed|DoGreen|DoBlue;
    Color.red  =Palet[0]*0x0101;
    Color.green=Palet[1]*0x0101;
    Color.blue =Palet[2]*0x0101;

    if ( (res=XAllocColor(Dsp,CMap,&Color)) ) {
      index=i;
      /* check if a previous allocation gave the same Color.pixel */
      for (j=0 ; (j<ncol) ; j++) 
	if (coltable[j].alocd && (coltable[j].colno==Color.pixel)) {
	  index=j;
	  XFreeColors(Dsp, CMap, &Color.pixel, 1, 0);
	  break;
	}
      if (index==i) {
	coltable[i].alocd=res;
	memcpy(coltable[i].ci_pal,Palet,3);
	coltable[i].colno=Color.pixel;
      }
      goto found;
    }
    else 
      i=ncol;
  }

  for(i=0, imin=mindist=-1 ; i<ncol ; i++)
    if ((coltable[i].alocd) &&
	((mindist<0) || (pPalDist(coltable[i].ci_pal,Palet) < mindist ))) {
      mindist=pPalDist(coltable[i].ci_pal,Palet);
      imin=i;
    }
  if (imin<0) {
    printf("shouldn't happen: Color Failure: R:%d G:%d B:%d \n",
	   Palet[0],Palet[1],Palet[2]);
    return Swap(black);
  }
  index=imin;
  memcpy(pal2,coltable[index].ci_pal,3);

found:
  *allind=index;
  coltable[index].refcnt++;
  return Swap(coltable[index].colno);
}

static ColTyp 
GetColTrCo(byte *Palet, int *allind, byte *pal2)
{
  byte Pal2p[3];
  int j;

  for(j=0 ; j<3 ; j++)
    Pal2p[j]=(Palet[j]>>3); /* May not be correct */

  return GetTrCol2p(Pal2p);
}

static ColTyp 
GetTrCol2p(byte *Palet)
{
  long tmp;
  XID res;
  int j;

  for(j=0,res=0 ; j<3 ; j++) {
    tmp=(Palet[j]*0x08421084L) >> shift[j];
    res|= (tmp & mask[j]);
  }
  return Swap(res); 
}

static ColTyp 
UseXPal8(byte *Palet)
{
  return XPal8[((Palet[1]<<3)&0xe0)
	      |(Palet[0]&0x1c)
	      |((Palet[2]>>3)&0x03)]; 
}

/* deallocate all allocated colors */
void 
TrashColor(void)
{
  int j;
  byte dummy[3];

  if (MutCMap)
    {
      for (j=0 ; j<16 ; j++)
	  UnGetColor(&alind[j],Pal[j]);
      
      for (j=0 ; j<256 ; j++)
	  UnGetColor(&alind8[j], dummy);
      
      /* check if we really deallocated all of them */
      for (j=nprealloc ; j < Ncoltable ; j++) 
	{
	  if (coltable[j].alocd)
	    {
	      XFreeColors(Dsp,CMap,&coltable[j].colno,1,0);	
	      coltable[j].alocd=0;
	    }
	  if (Verbose && coltable[j].refcnt) 
	    printf("Color for coltab[%d] is still %d times referenced!\n",
		   j,coltable[j].refcnt);
	}
    }
  if (CMapIsMine)
    XFreeColormap(Dsp,CMap);
}

void 
UnGetColor(int *ind, byte *pal)
{
    if (*ind>=0) {
      if (--(coltable[*ind].refcnt)==0) 
	{
	  coltable[*ind].cntdwn=2;
	}
      *ind=-1;
      /*      memcpy( pal, nopal, 3); prevent further use */     
    }
}

static void 
RefreshColors(void)
{ int i;
  for (i=nprealloc ; i<ncol ; i++)
    if ((coltable[i].refcnt==0) &&
	(coltable[i].alocd) &&
	(--coltable[i].cntdwn==0))
      {
	XFreeColors(Dsp,CMap,&coltable[i].colno,1,0);	
	coltable[i].alocd=0;
      }
}

/* Various auxilary routines */
static int 
SearchColor(byte *PalReq)
{
  int i,found;
  for (i=0, found=-1 ; i<ncol ; i++) {
      if (coltable[i].alocd && (!pPalCmp(PalReq, coltable[i].ci_pal)))
	{
	  found=i; break;
	}
  }
  return found;
}

static int 
ColCmp(byte *palx, byte *paly)
{
  return memcmp(palx,paly,3);
}

static int
ColDist(byte * pal1, byte *pal2)
{
    int dr,dg,db;

    dr=((pal1[0]>>5)&7)-((pal2[0]>>5)&7);
    dg=((pal1[1]>>5)&7)-((pal2[1]>>5)&7);
    db=((pal1[2]>>5)&7)-((pal2[2]>>5)&7);

    return (2*dr*dr+4*dg*dg+db*db);
}

static int 
CompareY(byte *palx, byte *paly)
{
  return Y(palx)!=Y(paly);
}

static int
GrayDist(byte * pal1, byte *pal2)
{
  return abs(Y(pal1)-Y(pal2));
}

static void 
SetColorStat(byte N,byte R,byte G,byte B)
{
  byte *pal;

  pal=Palette[N];
  pal[0]=R;
  pal[1]=G;
  pal[2]=B;
  XPal[N]=pGetColor(Palette[N],&alind[N],Pal[N]); 
}

static void 
SetColorMut(byte N,byte R,byte G,byte B)
{
  byte *pal;

  pal=Palette[N];
  pal[0]=R;
  pal[1]=G;
  pal[2]=B;
  if (pPalCmp(Palette[N],Pal[N]))
    {
      UnGetColor(&alind[N],Pal[N]);
      if ((PalOut&(1<<N))==0)
	XPal[N]=pGetColor(Palette[N],&alind[N],Pal[N]); 
    }
}

static void 
TakeColors(int ColSet)
{
  int i,msk;

  for ( i=0,msk=1 ; i<16 ; i++,msk<<=1)
    if (ColSet & msk & PalOut)
      {
	 XPal[i]=pGetColor(Palette[i],&alind[i],Pal[i]);
	 PalOut &= ~msk;
      }
}

void 
LeaveColors(int ColSet)
{
  int i,msk;

  for ( i=0,msk=1 ; i<16 ; i++,msk<<=1)
    if (ColSet & msk & ~PalOut)
      {
	 UnGetColor(&alind[i],Pal[i]);
	 PalOut |= msk;
      }
}

static void 
TakeColorsScr8(void)
{
  int I,J;
  byte Palet[3],Dummy[3];

  if (IsPal8) 
    return;
/* Allocate 'true colors' for screen 8 */
  for (I=255 ; I>=0 ; I--)
  {
    /* The X-server may run out of colormap entries so we mix up bits
       to spread the colors we can get: */
    J=((I&0x01)<<7)|((I&0x0A)<<3)|((I&0x54)>>1)|((I&0xA0)>>5);

    Palet[0]=(J&0x1C)<<3; /* red */
    Palet[1]=J&0xE0; /* green */
    Palet[2]=((J&0x03)==3) ? (7<<5) : ((J&0x03)<<6); /* blue */
    XPal8[J]=pGetColor(Palet,&alind8[J],Dummy);
  }
  if (Verbose&0x20) printf("Got colors for screen 8.\n");
  IsPal8=1;
}

static void 
LeaveColorsScr8(void)
{
  int J;
  byte dummy[3];

  if (!IsPal8)
    return;
/* release colors from screen 8 if necessary: */
  for (J=0; J<256; J++)
      UnGetColor(&alind8[J], dummy);

   IsPal8=0;
}



